from .db_utils import *
##TABLE_NAMES = get_all_table_names()
##TENANTS = fetch_any_combo(tableName='tenants')
##DOCUMENTS = fetch_any_combo(tableName='documents')
##PAGES = fetch_any_combo(tableName='pages')
##DOCUMENT_SEO = fetch_any_combo(tableName="document_seo")
##ANALYSIS_TYPES = fetch_any_combo(tableName="analysis_types")
##PAGE_ANALYSIS = fetch_any_combo(tableName="page_analysis")
##DOCUMENT_TAGS = fetch_any_combo(tableName="document_tags")
##TAGS = fetch_any_combo(tableName="tags")
##ANALYSIS_STRING = ""
##for ANALYSIS_TYPE in ANALYSIS_TYPES:
##    ANALYSIS_STRING+=str(ANALYSIS_TYPE)+'\n'
##STATS = f"""
##tables: {TABLE_NAMES}
##
##tenants: {TENANTS}
##
##analysis_types: {ANALYSIS_STRING}
##
##pages: {len(PAGES)}
##page_analysis: {len(PAGE_ANALYSIS)}
##SAMPLE: {PAGE_ANALYSIS[0]}
##documents: {len(DOCUMENTS)}
##document_seo: {len(DOCUMENT_SEO)}
##SAMPLE: {DOCUMENT_SEO[0]}
##
##tags: {TAGS}
##document_tags: {len(DOCUMENT_TAGS)}
##"""
##print(STATS)



from .imports import *
def get_item_vars(item_type=None,search_strings=None):
    matching = []
    data = get_data()
    if search_strings != None:
        search_strings = make_list(search_strings)
    items = data.get(str(item_type))#[item for item in data if item.get('kind') == str(item_type)]
    if not item_type or not items:
        images = data.get('images')
        pdfs = data.get('pdfs')
        items = pdfs+images
    for item in items:
        
        if search_strings:
            
            if is_search_strings(item,search_strings):
                matching.append(item)
        else:
            matching.append(item)
    
    return matching
def fix_jsons():
    dirs,files = get_files_and_dirs(IMGS_DIR,allowed_exts=['.json'])
    files = [file for file in files if file.endswith('info.json')]
    for file in files:
        
        directory = os.path.dirname(file)
        get_images_from_dir(directory)
        index_path = os.path.join(directory,'index.html')
def write_image_htmls(fix_paths = False):
    site = SiteConfig(
        base_url=ROOT_URL,
        site_name=SITE_NAME,
        domain=SITE_NAME,
        site_root=ROOT_URL,
        media_root=MEDIA_DIR,
        output_root=MEDIA_DIR,
        pdfs_dir=PDFS_DIR,
        imgs_dir=IMGS_DIR,
        videos_dir=VIDEOS_DIR,
    )
    env  = make_env(IMAGE_HTML_TEMPLATES)
    reg  = build_registry(MEDIA_DIR)
    for rel_dir, entry in reg.items():
        directory = os.path.join(MEDIA_DIR, rel_dir)
        info_path = os.path.join(directory,'info.json')
        if fix_paths:
            directory = fix_path(info_path)
            info_path = os.path.join(directory,'info.json')
        rel_dir = str(directory).replace(SITE_ROOT,"")

        get_images_from_dir(directory)
        data = correct_urls(info_path) or entry
        out_dir = create_out_path(rel_dir)
        
        out_path = render_to_file(entry = entry, rel_dir = rel_dir, site=site, env=env,
                       out_dir=out_dir)
        if out_path:
            print(path_to_url(out_path))



def process_page(repo: Repository, document_id: int, page_number: int, page_dir: str):
    """
    Process a single page: extract text, analyze, generate metadata, save to DB.
    
    Args:
        repo: Repository instance (injected)
        document_id: Document ID from DB
        page_number: Page number (1-indexed)
        page_dir: Directory path containing page assets
    """
    
    # 1. Create/upsert page record with file paths
    image_path = os.path.join(page_dir, "image.png")
    text_path = os.path.join(page_dir, "text.txt")
    info_path = os.path.join(page_dir, "info.json")
    metadata_path = os.path.join(page_dir, "metadata.json")
    
    page = repo.upsert_page(
        document_id=document_id,
        page_number=page_number,
        image_path=image_path,
        text_path=text_path,
        info_path=info_path,
        metadata_path=metadata_path
    )
    
    # 2. Extract text
    if not os.path.isfile(text_path):
        text_content = extract_single_pdf_page_text(pdf_path, page_number)
        write_to_file(file_path=text_path, contents=text_content)
    else:
        text_content = read_from_file(text_path)
    
    # 3. Analyze page
    if not os.path.isfile(info_path):
        analysis = analyze_page(text_content)
        safe_dump_to_json(file_path=info_path, data=analysis)
    else:
        analysis = safe_load_from_json(info_path)
    
    # 4. Generate metadata
    if not os.path.isfile(metadata_path):
        metadata = generate_metadata(analysis, page_number, document_id)
        safe_dump_to_json(file_path=metadata_path, data=metadata)
    else:
        metadata = safe_load_from_json(metadata_path)
    
    # 5. Save everything to DB
    repo.save_page_content(
        page_id=page.id,
        text_content=text_content,
        analysis=analysis,
        metadata=metadata
    )
    
    print(f"Processed page {page_number}")

def analyze_page(text_content: str) -> dict:
    """Analyze page text, return analysis dict."""
    data = {
        "text": text_content,
        "scope": "page",
        "summary_preset": "brief",
        "keyword_preset": "seo",
        "input_policy": "allow"
    }
    return postRequest('https://clownworld.biz/hugpy/analyze/text', data=data)

def generate_metadata(analysis: dict, page_number: int, document_id: int) -> dict:
    """Generate metadata from analysis."""
    data = {"info": analysis}
    return postRequest('https://clownworld.biz/metadata/get/info', data=data)

# Main workflow
if __name__ == "__main__":
    conn = psycopg.connect(DATABASE_URL, autocommit=True)
    repo = Repository(conn)
    
    try:
        # Get pending documents
        pending_docs = repo.list_documents(TENANT_ID, status=Status.PENDING)
        
        for doc in pending_docs:
            print(f"Processing {doc.slug}...")
            pages_dir = os.path.join(doc.base_path, 'pages')
            
            if not os.path.isdir(pages_dir):
                print(f"  No pages directory")
                continue
            
            # Process each page
            for i in range(1, doc.page_count + 1):
                page_str = f"{i:04d}"
                page_dir = os.path.join(pages_dir, page_str)
                
                if not os.path.isdir(page_dir):
                    continue
                
                process_page(repo, doc.id, i, page_dir)
            
            # Mark document complete
            repo.update_document_status(doc.id, Status.COMPLETE)
    
    finally:
        conn.close()



# ── Schemas ──────────────────────────────────────────────────
from .imports import *

class Status(str, Enum):
    PENDING   = "pending"
    INGESTING = "ingesting"
    ANALYZING = "analyzing"
    COMPLETE  = "complete"
    FAILED    = "failed"


@dataclass(frozen=True, slots=True)
class Tenant:
    id: str
    name: str
    slug: str
    created_at: datetime


@dataclass(frozen=True, slots=True)
class Document:
    id: int
    tenant_id: str
    doc_id: str
    slug: str
    base_path: str
    pdf_path: str
    status: Status
    page_count: int | None
    created_at: datetime
    updated_at: datetime


@dataclass(frozen=True, slots=True)
class Page:
    id: int
    document_id: int
    page_number: int
    image_path: str | None
    text_path: str | None
    info_path: str | None
    metadata_path: str | None
    created_at: datetime
    ocr_processed: bool = False

# classes.py — replace the existing PageAnalysis dataclass with this

@dataclass(frozen=True, slots=True)
class PageAnalysis:
    id: int
    page_id: int
    analysis_type: str
    payload: dict
    model_version: str | None
    created_at: datetime

@dataclass(frozen=True, slots=True)
class PageMetadata:
    """Explicit schema for page metadata."""
    title: str
    description: str
    description_html: str
    keywords: str
    thumbnail: str | None
    canonical: str
    mobile_url: str
    og: dict  # type, title, description, image, etc.
    twitter: dict  # card, title, description, etc.
    other: dict  # robots, author, viewport, etc.


@dataclass(frozen=True, slots=True)
class PipelineRun:
    id: int
    document_id: int
    status: Status
    stage: str
    error_message: str | None
    started_at: datetime
    finished_at: datetime | None


@dataclass(frozen=True, slots=True)
class Tag:
    id: int
    slug: str
    label: str


# ── Identity hashing ────────────────────────────────────────


def identity_hash(slug: str, discriminator: str = "") -> str:
    """Deterministic document ID from stable domain keys."""
    return hashlib.sha256(f"{slug}:{discriminator}".encode()).hexdigest()



import os
from .imports import *
def write_atomic(path: str, content: str) -> None:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    tmp = f"{path}.{os.getpid()}.tmp"
    write_to_file(file_path = tmp,contents=content)
    os.replace(tmp, path)




from __future__ import annotations
import shutil,logging,json,os,fitz,re,sys,hashlib,argparse
logging.getLogger("pydot").setLevel(logging.WARNING)
logging.getLogger("pypdf").setLevel(logging.WARNING)
logging.getLogger("paddle").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import *
from enum import Enum
from jinja2 import (
    Environment,
    FileSystemLoader,
    select_autoescape
    )







from .imports import *
from .template_registry import *
from .assembly import *
from .helpers import *
from .rendering import *
# ═══════════════════════════════════════════════════════════════
# Directory walker — the recursive orchestrator
# ═══════════════════════════════════════════════════════════════


@dataclass
class WalkConfig:
    """All the explicit wiring a walk needs."""
    site_root: str
    dry_run: bool
    templates: TemplateRegistry
    resolver: PathResolver
    doc_index: DocumentIndex


def _write_html(path: Path, html: str, label: str, dry_run: bool) -> None:
    if dry_run:
        print(f"[dry-run] {path}  [{label}]")
    else:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(html, encoding="utf-8")
        print(f"wrote {path}  [{label}]")


def _load_page_metadata(page_dir: Path) -> dict:
    """
    Load metadata from a page directory.
    Merges info.json and metadata.json into a single dict.
    """
    metadata = {}
    
    # info.json has basic image/caption info
    info_path = page_dir / "info.json"
    if info_path.exists():
        try:
            metadata.update(json.loads(info_path.read_text()))
        except (json.JSONDecodeError, OSError) as e:
            logger.warning("Bad info.json in %s: %s", page_dir, e)
    
    # metadata.json has SEO-specific fields (title, description, keywords, etc.)
    meta_path = page_dir / "metadata.json"
    if meta_path.exists():
        try:
            metadata.update(json.loads(meta_path.read_text()))
        except (json.JSONDecodeError, OSError) as e:
            logger.warning("Bad metadata.json in %s: %s", page_dir, e)
    
    return metadata


def _generate_page_seo_page(
    page_dir: Path,
    page_num: int,
    document: Any,
    view: DocumentView,
    config: WalkConfig,
    base_url: str,
) -> None:
    """
    Generate individual SEO page for a page directory.
    
    Each page gets a dedicated HTML with:
    - Page image from image.png
    - OCR text from text.txt
    - SEO metadata from metadata.json + info.json
    - Structured schema for the page
    """
    out_file = page_dir / "index.html"
    
    # Load metadata
    metadata = _load_page_metadata(page_dir)
    
    # Resolve paths
    image_path = page_dir / "image.png"
    text_path = page_dir / "text.txt"
    
    image_url = config.resolver.to_url_if_exists(str(image_path)) if image_path.exists() else None
    ocr_text = read_text_file(str(text_path)) if text_path.exists() else ""
    
    # Extract SEO fields with sensible fallbacks
    page_title = metadata.get("title") or f"{view.title} — Page {page_num}"
    description = metadata.get("description") or clean_text(ocr_text or view.description, 300)
    keywords = metadata.get("keywords") or view.keywords
    alt_text = metadata.get("alt") or f"Page {page_num} of {view.title}"
    
    # Social metadata from info.json
    social_meta = metadata.get("social_meta", {})
    og_image = social_meta.get("og:image") or image_url
    
    # Build canonical URL for this page
    # Assuming base_url is the document dir, pages are at base_url/pages/XXXX/
    page_canonical = f"{base_url.rstrip('/')}/pages/{page_dir.name}/"
    index_url = base_url.rstrip("/") + "/"
    
    # Render the page
    html = render_page_seo(
        config.templates,
        page_title=page_title,
        description=description,
        keywords=keywords,
        alt_text=alt_text,
        image_url=image_url,
        ocr_text=ocr_text,
        page_num=page_num,
        document_title=view.title,
        document_keywords=view.keywords,
        canonical_url=page_canonical,
        og_image=og_image,
        page_url=page_canonical,
        document_url=index_url,
        site_root=config.site_root,
        schema_data=metadata.get("schema", {}),
    )
    
    _write_html(out_file, html, f"page SEO, page {page_num}", config.dry_run)


def _generate_pages_for_document(
    document_dir: Path,
    base_url: str,
    document: Any,
    view: DocumentView,
    config: WalkConfig,
) -> None:
    """
    Generate SEO pages for all pages under pages/ subdirectory.
    
    Looks for document_dir/pages/XXXX/ directories and renders each one.
    """
    pages_dir = document_dir / "pages"
    if not pages_dir.exists():
        return
    
    # Find all page directories (0001, 0002, etc.)
    page_dirs = sorted(
        d for d in pages_dir.iterdir()
        if d.is_dir() and not d.name.startswith(".")
    )
    
    for i, page_dir in enumerate(page_dirs, 1):
        _generate_page_seo_page(
            page_dir=page_dir,
            page_num=i,
            document=document,
            view=view,
            config=config,
            base_url=base_url,
        )


def generate_for_directory(
    directory: Path,
    base_url: str,
    config: WalkConfig,
    recurse: bool = True,
) -> bool:
    """
    Generate index.html for this directory. Returns True if something was written.

    Decision order:
      1. Does this dir match a DB document with SEO? → viewer page + page SEO pages
      2. Does this dir have a filesystem manifest?   → viewer page (filesystem fallback)
      3. Does this dir have info.json?               → image page
      4. Does this dir have children?                → gallery page
      5. Otherwise                                   → skip
    """
    out_file = directory / "index.html"

    # ── 1. DB document match → viewer page ──
    doc = config.doc_index.find_by_dir(directory)
    if doc:
        view = config.doc_index.assemble_view(
            doc.id, config.resolver, base_url
        )
        if view and view.pages:
            html = render_viewer(config.templates, view, config.site_root,directory)
            _write_html(out_file, html, f"viewer, {len(view.pages)} pages", config.dry_run)
            
            # Generate individual page SEO pages (pages/XXXX/index.html)
            _generate_pages_for_document(
                document_dir=directory,
                base_url=base_url,
                document=doc,
                view=view,
                config=config,
            )
            
            # Viewer pages are leaves — don't recurse into their subdirs
            return True

    # ── 2. Filesystem manifest (not yet in DB) → viewer page ──
    # Preserves v5 behavior for dirs that haven't been ingested yet
    manifest = load_manifest(directory)
    if manifest:
        html = _render_manifest_viewer(
            directory, base_url, manifest, config
        )
        _write_html(out_file, html, f"viewer/manifest, {len(manifest)} pages", config.dry_run)
        return True

    # ── 3. info.json → image page ──
    info_path = directory / "info.json"
    if info_path.exists():
        try:
            meta = json.loads(info_path.read_text())
            html = render_image_page(
                config.templates, meta, base_url,
                config.site_root, directory, config.resolver,
            )
            _write_html(out_file, html, "image page", config.dry_run)
            return True
        except (json.JSONDecodeError, OSError) as e:
            logger.warning("Bad info.json %s: %s", info_path, e)

    # ── 4. Branch dir with children → gallery ──
    children = child_dirs(directory)
    if not children:
        return False

    # Recurse first so child pages exist before we build the gallery
    if recurse:
        for child in children:
            child_url = f"{base_url.rstrip('/')}/{child.name}"
            generate_for_directory(child, child_url, config, recurse=True)

    # Build gallery cards for immediate children
    cards = [
        config.doc_index.make_card_for_dir(
            child,
            href=f"{base_url.rstrip('/')}/{child.name}/",
            resolver=config.resolver,
        )
        for child in children
    ]

    html = render_gallery(
        config.templates, cards, humanize(directory.name),
        base_url, config.site_root,
    )
    _write_html(out_file, html, f"gallery, {len(cards)} cards", config.dry_run)
    return True


def _render_manifest_viewer(
    directory: Path,
    base_url: str,
    manifest: list[dict],
    config: WalkConfig,
) -> str:
    """
    Render a viewer page from a filesystem manifest (not yet in DB).
    This is the backward-compat path for dirs that haven't been ingested.
    Uses the same viewer template, just assembles context differently.
    """
    title = humanize(directory.name)
    first = manifest[0]
    description = clean_text(
        first.get("longdesc") or first.get("caption") or "", 300
    )
    keywords_raw = set()
    for entry in manifest:
        for kw in entry.get("keywords_str", "").split(","):
            kw = kw.strip()
            if kw and len(kw) > 2:
                keywords_raw.add(kw)
    keywords = ", ".join(sorted(keywords_raw)[:30])
    keywords_list = sorted(keywords_raw)[:12]

    pages_js = []
    for i, entry in enumerate(manifest, 1):
        thumb = entry.get("social_meta", {}).get("og:image")
        pages_js.append({
            "n": i,
            "thumb": thumb or "",
            "text": entry.get("longdesc", "").strip(),
            "alt": entry.get("alt", "").split(" | ")[0] or f"Page {i}",
        })

    pdf_url = first.get("schema", {}).get("url", "")
    first_thumb = pages_js[0]["thumb"] if pages_js else ""

    schema = {
        "@context": "https://schema.org",
        "@type": "CreativeWork",
        "name": title,
        "description": description,
        "keywords": keywords,
        "url": pdf_url,
        "thumbnailUrl": first_thumb,
        "fileFormat": "application/pdf",
    }

    return config.templates.render(
        "viewer",
        title=title,
        description=description,
        keywords=keywords,
        canonical_url=base_url.rstrip("/") + "/",
        first_thumb=first_thumb,
        schema_json=json.dumps(schema, indent=2),
        site_root=config.site_root,
        total=len(pages_js),
        pdf_url=pdf_url,
        pages_json=json.dumps(pages_js, ensure_ascii=False),
        pdf_url_json=json.dumps(pdf_url),
        keywords_json=json.dumps(keywords_list),
    )



from pathlib import Path
from .upload_document import upload_document
from .imports import *
import json
from abstract_utilities import get_any_value
ROOT = Path("/srv/staging/pdfs")
def check_keywords_summary(data):
    keywords = data.get("keywords")
    if not keywords:
        return False
    summary = data.get("summary")
    description = data.get("description")
    if not summary and not description:
        return False
    return True    
def get_key_value_from_data(data,key):
    value = data.get(key)
    if not value:
        values = get_any_value(data,key)
        if not values:
            return value 
        values = [value for value in values if value]
        if values:
            value = values[0]
    return value
def get_analysis_from_info(info_path):
    data_keys = {}
    info = safe_load_from_json(str(info_path))
    analysis_keys = ["alt","caption","keyword","keywords_str","summary","description","title","longdesc"]
    for key in analysis_keys:
        value = get_key_value_from_data(info,key)
        if value and key in ["summary","description"] and not data_keys.get("summary"):
            data_keys["summary"] = value
    return data_keys
def find_documents(root: Path):
    for p in root.rglob("pages"):
        yield p.parent


def get_best_text(base_dir: Path):
    """
    Pick best available text file (first found for now)
    """
    for txt in base_dir.rglob("text.txt"):
        return txt
    return None
def process_pages(base_dir: Path, doc_id: int):
    pages_dir = base_dir / "pages"

    for page_dir in sorted(pages_dir.glob("page_*")):
        text_file = page_dir / "text.txt"
        info_file = page_dir / "info.json"

        analysis = None

        # -------------------------
        # 1. Prefer existing info.json
        # -------------------------
        if info_file.exists():
##                try:
                info = get_analysis_from_info(info_file)
                if check_keywords_summary(info):
                    analysis = info
                    print(f"📦 Using existing info.json: {page_dir}")

##                except Exception as e:
##                    print(f"⚠️ Invalid info.json, falling back: {page_dir}")
##                    print(e)

        # -------------------------
        # 2. Fallback → analyze
        # -------------------------
        if analysis is None:
            if not text_file.exists():
                continue
               
            analysis = analyze_page(str(text_file))

            if not analysis:
                continue
    
            print(f"🧠 Analyzed: {page_dir}")
            
            # OPTIONAL: persist back to info.json
            try:
                with open(info_file, "w") as f:
                    json.dump(analysis, f, indent=2)
            except Exception as e:
                print(f"⚠️ Failed to write info.json: {page_dir}")
                print(e)

        # -------------------------
        # 3. Insert into DB
        # -------------------------
        insert_page_analysis(doc_id, page_dir, analysis)

##        except Exception as e:
##            print(f"❌ Page failed: {page_dir}")
##            print(e)

def process_document(base_dir: Path):
    print(f"📄 Processing: {base_dir}")

    base_path = str(base_dir)

    if document_exists_by_path(base_path):
        print("⏭️ Skipping (already exists)")
        return

    # ✅ Step 1: insert document ONLY
    doc_id = upload_document(base_dir)

    if not doc_id:
        print("⏭️ Already exists")
        return

    print(f"🧱 Document inserted: {doc_id}")

    # ✅ Step 2: insert pages (already happens inside upload_document)

    # ✅ Step 3: analyze pages (independent)
    process_pages(base_dir, doc_id)

    print(f"✅ Completed document {doc_id}")



#!/usr/bin/env python3
"""
generate_index_db.py — Static HTML generation from the document registry DB.

Walks the filesystem tree for directory structure, queries the DB for content,
and renders through Jinja2 templates. Every directory gets an index.html:
  - Leaf dirs matching a DB document → viewer page
  - Leaf dirs with info.json (no DB match) → image page
  - Branch dirs with children → gallery index

Design:
  - Filesystem provides structure (which directories exist).
  - DB provides content (titles, descriptions, SEO, page data).
  - Jinja2 templates own all HTML. No f-string markup in Python.
  - Templates are loaded from a registry, not discovered at runtime.
  - All config is explicit. No smart defaults.

Usage:
    python generate_index_db.py \\
        --dsn           "postgresql://user:pass@localhost:5432/tdd_docs" \\
        --tenant-slug   default \\
        --root          /srv/media/thedailydialectics/pdfs \\
        --media-root    /srv/media/thedailydialectics \\
        --base-url      https://thedailydialectics.com/pdfs \\
        --site-root     https://thedailydialectics.com

    # Single directory:
    python generate_index_db.py \\
        --dsn           "postgresql://user:pass@localhost:5432/tdd_docs" \\
        --tenant-slug   default \\
        --root          /srv/media/thedailydialectics/pdfs/wipow/a197278 \\
        --media-root    /srv/media/thedailydialectics \\
        --base-url      https://thedailydialectics.com/pdfs/wipow/a197278 \\
        --site-root     https://thedailydialectics.com \\
        --no-recurse --dry-run
"""

from __future__ import annotations

import argparse
import json
import logging
import os
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import psycopg
from jinja2 import Environment, FileSystemLoader, select_autoescape

from repository import Repository, Status
from repository_seo import SeoRepository, DocumentSeo

logger = logging.getLogger(__name__)


# ═══════════════════════════════════════════════════════════════
# Constants
# ═══════════════════════════════════════════════════════════════

SITE_ROOT_DEFAULT = "https://thedailydialectics.com"

SKIP_DIRS = frozenset({
    "text", "pages", "pdf_pages",
    "preprocessed_images", "preprocessed_text",
    "node_modules", ".git", "__pycache__",
})

IMAGE_EXTS = frozenset({".webp", ".jpg", ".jpeg", ".png", ".gif"})


# ═══════════════════════════════════════════════════════════════
# Intermediate shapes
# ═══════════════════════════════════════════════════════════════


@dataclass(frozen=True, slots=True)
class PageView:
    page_number: int
    image_url: str | None
    text: str
    alt: str


@dataclass(frozen=True, slots=True)
class DocumentView:
    document_id: int
    slug: str
    title: str
    description: str
    keywords: str
    pdf_url: str
    thumbnail_url: str | None
    canonical_url: str
    tags: list[str]
    pages: list[PageView]


@dataclass(frozen=True, slots=True)
class GalleryCard:
    title: str
    description: str
    image_url: str | None
    href: str
    tags: list[str]
    page_count: int | None


@dataclass(frozen=True, slots=True)
class ImagePageView:
    title: str
    alt: str
    description: str
    keywords: str
    keyword_tags: list[str]
    img_url: str | None
    og_image: str
    schema_json: str
    license: str
    attribution: str
    canonical_url: str
    breadcrumbs: str


# ═══════════════════════════════════════════════════════════════
# Path ↔ URL resolution
# ═══════════════════════════════════════════════════════════════


@dataclass(frozen=True, slots=True)
class PathResolver:
    media_root: Path
    site_root: str

    def to_url(self, path: str | Path) -> str | None:
        p = Path(path)
        if not p.is_absolute():
            return None
        try:
            return f"{self.site_root}/{p.relative_to(self.media_root)}"
        except ValueError:
            return None

    def to_url_if_exists(self, path: str | Path | None) -> str | None:
        if path is None:
            return None
        p = Path(path)
        if p.exists():
            return self.to_url(p)
        return None


# ═══════════════════════════════════════════════════════════════
# Template registry
# ═══════════════════════════════════════════════════════════════


class TemplateRegistry:
    """
    Loads and caches Jinja2 templates from a directory.
    Templates are registered by name, not discovered at runtime.
    """

    KNOWN_TEMPLATES = frozenset({"viewer", "gallery", "image_page"})

    def __init__(self, template_dir: Path) -> None:
        if not template_dir.is_dir():
            raise FileNotFoundError(f"Template directory not found: {template_dir}")

        self._env = Environment(
            loader=FileSystemLoader(str(template_dir)),
            autoescape=select_autoescape(default_for_string=True),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        # Validate all known templates exist at init time, not render time
        for name in self.KNOWN_TEMPLATES:
            filename = f"{name}.html.j2"
            try:
                self._env.get_template(filename)
            except Exception as exc:
                raise FileNotFoundError(
                    f"Required template '{filename}' not found in {template_dir}"
                ) from exc

    def render(self, template_name: str, **ctx: Any) -> str:
        if template_name not in self.KNOWN_TEMPLATES:
            raise ValueError(
                f"Unknown template '{template_name}'. "
                f"Known: {sorted(self.KNOWN_TEMPLATES)}"
            )
        tmpl = self._env.get_template(f"{template_name}.html.j2")
        return tmpl.render(**ctx)


# ═══════════════════════════════════════════════════════════════
# Text helpers
# ═══════════════════════════════════════════════════════════════


def clean_text(s: str | list, max_len: int = 160) -> str:
    if isinstance(s, list):
        s = str(s[0]) if s else ""
    s = re.sub(r"\s+", " ", s).strip()
    if len(s) > max_len:
        s = s[:max_len].rsplit(" ", 1)[0] + "…"
    return s


def humanize(name: str) -> str:
    return name.replace("-", " ").replace("_", " ").title()


def breadcrumbs_html(base_url: str, site_root: str) -> str:
    path_part = base_url.rstrip("/").replace(site_root, "").lstrip("/")
    segments = [s for s in path_part.split("/") if s]
    crumbs = [f'<a href="{site_root}">Home</a>']
    acc = site_root
    for i, seg in enumerate(segments):
        acc += f"/{seg}"
        if i < len(segments) - 1:
            crumbs.append(f'<a href="{acc}/">{humanize(seg)}</a>')
        else:
            crumbs.append(f"<span>{seg}</span>")
    return " › ".join(crumbs)


def read_text_file(path: str | None) -> str:
    if not path:
        return ""
    p = Path(path)
    if not p.exists():
        return ""
    try:
        return p.read_text(encoding="utf-8", errors="replace").strip()
    except OSError:
        return ""


def child_dirs(directory: Path) -> list[Path]:
    return sorted(
        d for d in directory.iterdir()
        if d.is_dir() and d.name not in SKIP_DIRS and not d.name.startswith(".")
    )


def first_image_url(directory: Path, resolver: PathResolver) -> str | None:
    for ext in IMAGE_EXTS:
        for hit in sorted(directory.rglob(f"*{ext}")):
            if hit.is_file():
                url = resolver.to_url(hit)
                if url:
                    return url
    return None


def load_manifest(directory: Path) -> list[dict] | None:
    """Load a manifest file from the directory (filesystem fallback)."""
    for name in [f"{directory.name}_manifest.json", "manifest.json"]:
        p = directory / name
        if not p.exists():
            continue
        try:
            data = json.loads(p.read_text())
            if isinstance(data, list) and data:
                return data
        except (json.JSONDecodeError, OSError) as e:
            logger.warning("Bad manifest %s: %s", p, e)
    return None


# ═══════════════════════════════════════════════════════════════
# DB → intermediate shape assembly
# ═══════════════════════════════════════════════════════════════


class DocumentIndex:
    """
    Pre-loaded index of all documents for a tenant, keyed by slug
    and by base_path. Built once, queried many times during the walk.
    """

    def __init__(
        self,
        repo: Repository,
        seo_repo: SeoRepository,
        tenant_id: str,
    ) -> None:
        self._repo = repo
        self._seo_repo = seo_repo
        self._tenant_id = tenant_id

        # Load everything once
        all_docs = repo.list_documents(tenant_id, limit=100_000)
        self._by_slug: dict[str, Any] = {d.slug: d for d in all_docs}
        self._by_base_path: dict[str, Any] = {d.base_path: d for d in all_docs}

    def find_by_dir(self, directory: Path) -> Any | None:
        """Match a filesystem directory to a document, by slug or base_path."""
        # Try slug match first (most common: dir name == document slug)
        doc = self._by_slug.get(directory.name)
        if doc:
            return doc
        # Try base_path match
        return self._by_base_path.get(str(directory))

    def assemble_view(
        self, document_id: int, resolver: PathResolver, base_url: str,
    ) -> DocumentView | None:
        doc = self._repo.get_document(document_id)
        if doc is None:
            return None

        seo: DocumentSeo | None = self._seo_repo.get_seo(document_id)
        if seo is None:
            return None

        db_pages = self._repo.get_pages(document_id)
        page_metas = self._seo_repo.get_all_page_metas(document_id)
        meta_by_index = {i: m for i, m in enumerate(page_metas)}

        tags = [t.label for t in self._repo.get_document_tags(document_id)]

        pages: list[PageView] = []
        for i, pg in enumerate(db_pages):
            meta = meta_by_index.get(i) or {}
            image_url = resolver.to_url_if_exists(pg.image_path)
            text = read_text_file(pg.text_path)
            if not text:
                text = meta.get("longdesc", "") or meta.get("text", "")
            alt = (
                meta.get("alt")
                or meta.get("title")
                or f"{seo.title} — Page {pg.page_number}"
            )
            pages.append(PageView(
                page_number=pg.page_number,
                image_url=image_url,
                text=text,
                alt=alt,
            ))

        pdf_url = resolver.to_url(doc.pdf_path) or doc.pdf_path
        thumbnail = seo.thumbnail_url
        if not thumbnail and pages:
            thumbnail = pages[0].image_url

        return DocumentView(
            document_id=doc.id,
            slug=doc.slug,
            title=seo.title,
            description=seo.description,
            keywords=seo.keywords,
            pdf_url=pdf_url,
            thumbnail_url=thumbnail,
            canonical_url=seo.canonical_url or f"{base_url.rstrip('/')}/",
            tags=tags,
            pages=pages,
        )

    def make_card_for_dir(
        self, directory: Path, href: str, resolver: PathResolver,
    ) -> GalleryCard:
        """
        Build a gallery card for a child directory.
        Tries DB first, then filesystem manifest, then bare directory name.
        """
        doc = self.find_by_dir(directory)
        if doc:
            seo = self._seo_repo.get_seo(doc.id)
            if seo:
                thumb = seo.thumbnail_url
                if not thumb:
                    pages = self._repo.get_pages(doc.id)
                    if pages and pages[0].image_path:
                        thumb = resolver.to_url_if_exists(pages[0].image_path)
                tags = [t.label for t in self._repo.get_document_tags(doc.id)]
                return GalleryCard(
                    title=seo.title,
                    description=clean_text(seo.description, 160),
                    image_url=thumb,
                    href=href,
                    tags=tags,
                    page_count=doc.page_count,
                )

        # Filesystem fallback: manifest
        manifest = load_manifest(directory)
        if manifest:
            first = manifest[0]
            thumb_raw = first.get("social_meta", {}).get("og:image")
            return GalleryCard(
                title=humanize(directory.name),
                description=clean_text(
                    first.get("longdesc")
                    or first.get("caption")
                    or first.get("keywords_str", "").replace(",", " ")
                    or "",
                    160,
                ),
                image_url=thumb_raw,
                href=href,
                tags=[],
                page_count=len(manifest),
            )

        # Filesystem fallback: info.json
        info_path = directory / "info.json"
        if info_path.exists():
            try:
                meta = json.loads(info_path.read_text())
                raw_url = (
                    meta.get("schema", {}).get("url")
                    or meta.get("social_meta", {}).get("og:image")
                )
                return GalleryCard(
                    title=meta.get("title") or humanize(directory.name),
                    description=clean_text(
                        meta.get("longdesc") or meta.get("caption") or "", 160
                    ),
                    image_url=raw_url,
                    href=href,
                    tags=[],
                    page_count=None,
                )
            except (json.JSONDecodeError, OSError):
                pass

        # Bare directory — no metadata available
        return GalleryCard(
            title=humanize(directory.name),
            description="",
            image_url=first_image_url(directory, resolver),
            href=href,
            tags=[],
            page_count=None,
        )


# ═══════════════════════════════════════════════════════════════
# Rendering — templates + context
# ═══════════════════════════════════════════════════════════════


def render_viewer(
    templates: TemplateRegistry,
    view: DocumentView,
    site_root: str,
) -> str:
    pages_js = [
        {
            "n": pv.page_number,
            "thumb": pv.image_url or "",
            "text": pv.text,
            "alt": pv.alt.split(" | ")[0] if pv.alt else f"Page {pv.page_number}",
        }
        for pv in view.pages
    ]
    keywords_list = [
        kw.strip()
        for kw in view.keywords.split(",")
        if kw.strip() and len(kw.strip()) > 2
    ][:12]
    schema = {
        "@context": "https://schema.org",
        "@type": "CreativeWork",
        "name": view.title,
        "description": clean_text(view.description, 300),
        "keywords": view.keywords,
        "url": view.pdf_url,
        "thumbnailUrl": view.thumbnail_url or "",
        "fileFormat": "application/pdf",
    }

    return templates.render(
        "viewer",
        title=view.title,
        description=clean_text(view.description, 300),
        keywords=view.keywords,
        canonical_url=view.canonical_url,
        first_thumb=pages_js[0]["thumb"] if pages_js else "",
        schema_json=json.dumps(schema, indent=2),
        site_root=site_root,
        total=len(view.pages),
        pdf_url=view.pdf_url,
        pages_json=json.dumps(pages_js, ensure_ascii=False),
        pdf_url_json=json.dumps(view.pdf_url),
        keywords_json=json.dumps(keywords_list),
    )


def render_gallery(
    templates: TemplateRegistry,
    cards: list[GalleryCard],
    title: str,
    base_url: str,
    site_root: str,
) -> str:
    canonical = base_url.rstrip("/") + "/"
    card_dicts = [
        {
            "title": c.title,
            "description": c.description,
            "image_url": c.image_url,
            "href": c.href,
            "tags": c.tags,
            "page_count": c.page_count,
        }
        for c in cards
    ]
    return templates.render(
        "gallery",
        page_title=title,
        canonical_url=canonical,
        breadcrumbs=breadcrumbs_html(base_url, site_root),
        heading=title,
        card_count=len(cards),
        cards=card_dicts,
    )


def render_image_page(
    templates: TemplateRegistry,
    meta: dict[str, Any],
    base_url: str,
    site_root: str,
    directory: Path,
    resolver: PathResolver,
) -> str:
    title = meta.get("title") or humanize(directory.name)
    alt = meta.get("alt") or title
    description = clean_text(
        meta.get("longdesc") or meta.get("caption") or "", 300
    )
    keywords = meta.get("keywords_str", "")
    img_url = (
        meta.get("schema", {}).get("url")
        or meta.get("social_meta", {}).get("og:image")
    )
    # Filesystem fallback for image
    if not img_url:
        img_url = first_image_url(directory, resolver)

    keyword_tags = [kw.strip() for kw in keywords.split(",") if kw.strip()]

    return templates.render(
        "image_page",
        title=title,
        alt=alt,
        description=description,
        keywords=keywords,
        keyword_tags=keyword_tags,
        img_url=img_url,
        og_image=img_url or "",
        schema_json=json.dumps(meta.get("schema", {}), indent=2),
        license=meta.get("license", ""),
        attribution=meta.get("attribution", ""),
        canonical_url=base_url.rstrip("/") + "/",
        breadcrumbs=breadcrumbs_html(base_url, site_root),
    )


# ═══════════════════════════════════════════════════════════════
# Directory walker — the recursive orchestrator
# ═══════════════════════════════════════════════════════════════


@dataclass
class WalkConfig:
    """All the explicit wiring a walk needs."""
    site_root: str
    dry_run: bool
    templates: TemplateRegistry
    resolver: PathResolver
    doc_index: DocumentIndex


def _write_html(path: Path, html: str, label: str, dry_run: bool) -> None:
    if dry_run:
        print(f"[dry-run] {path}  [{label}]")
    else:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(html, encoding="utf-8")
        print(f"wrote {path}  [{label}]")


def generate_for_directory(
    directory: Path,
    base_url: str,
    config: WalkConfig,
    recurse: bool = True,
) -> bool:
    """
    Generate index.html for this directory. Returns True if something was written.

    Decision order:
      1. Does this dir match a DB document with SEO? → viewer page
      2. Does this dir have a filesystem manifest?   → viewer page (filesystem fallback)
      3. Does this dir have info.json?               → image page
      4. Does this dir have children?                → gallery page
      5. Otherwise                                   → skip
    """
    out_file = directory / "index.html"

    # ── 1. DB document match → viewer page ──
    doc = config.doc_index.find_by_dir(directory)
    if doc:
        view = config.doc_index.assemble_view(
            doc.id, config.resolver, base_url
        )
        if view and view.pages:
            html = render_viewer(config.templates, view, config.site_root)
            _write_html(out_file, html, f"viewer, {len(view.pages)} pages", config.dry_run)
            # Viewer pages are leaves — don't recurse into their subdirs
            return True

    # ── 2. Filesystem manifest (not yet in DB) → viewer page ──
    # Preserves v5 behavior for dirs that haven't been ingested yet
    manifest = load_manifest(directory)
    if manifest:
        html = _render_manifest_viewer(
            directory, base_url, manifest, config
        )
        _write_html(out_file, html, f"viewer/manifest, {len(manifest)} pages", config.dry_run)
        return True

    # ── 3. info.json → image page ──
    info_path = directory / "info.json"
    if info_path.exists():
        try:
            meta = json.loads(info_path.read_text())
            html = render_image_page(
                config.templates, meta, base_url,
                config.site_root, directory, config.resolver,
            )
            _write_html(out_file, html, "image page", config.dry_run)
            return True
        except (json.JSONDecodeError, OSError) as e:
            logger.warning("Bad info.json %s: %s", info_path, e)

    # ── 4. Branch dir with children → gallery ──
    children = child_dirs(directory)
    if not children:
        return False

    # Recurse first so child pages exist before we build the gallery
    if recurse:
        for child in children:
            child_url = f"{base_url.rstrip('/')}/{child.name}"
            generate_for_directory(child, child_url, config, recurse=True)

    # Build gallery cards for immediate children
    cards = [
        config.doc_index.make_card_for_dir(
            child,
            href=f"{base_url.rstrip('/')}/{child.name}/",
            resolver=config.resolver,
        )
        for child in children
    ]

    html = render_gallery(
        config.templates, cards, humanize(directory.name),
        base_url, config.site_root,
    )
    _write_html(out_file, html, f"gallery, {len(cards)} cards", config.dry_run)
    return True


def _render_manifest_viewer(
    directory: Path,
    base_url: str,
    manifest: list[dict],
    config: WalkConfig,
) -> str:
    """
    Render a viewer page from a filesystem manifest (not yet in DB).
    This is the backward-compat path for dirs that haven't been ingested.
    Uses the same viewer template, just assembles context differently.
    """
    title = humanize(directory.name)
    first = manifest[0]
    description = clean_text(
        first.get("longdesc") or first.get("caption") or "", 300
    )
    keywords_raw = set()
    for entry in manifest:
        for kw in entry.get("keywords_str", "").split(","):
            kw = kw.strip()
            if kw and len(kw) > 2:
                keywords_raw.add(kw)
    keywords = ", ".join(sorted(keywords_raw)[:30])
    keywords_list = sorted(keywords_raw)[:12]

    pages_js = []
    for i, entry in enumerate(manifest, 1):
        thumb = entry.get("social_meta", {}).get("og:image")
        pages_js.append({
            "n": i,
            "thumb": thumb or "",
            "text": entry.get("longdesc", "").strip(),
            "alt": entry.get("alt", "").split(" | ")[0] or f"Page {i}",
        })

    pdf_url = first.get("schema", {}).get("url", "")
    first_thumb = pages_js[0]["thumb"] if pages_js else ""

    schema = {
        "@context": "https://schema.org",
        "@type": "CreativeWork",
        "name": title,
        "description": description,
        "keywords": keywords,
        "url": pdf_url,
        "thumbnailUrl": first_thumb,
        "fileFormat": "application/pdf",
    }

    return config.templates.render(
        "viewer",
        title=title,
        description=description,
        keywords=keywords,
        canonical_url=base_url.rstrip("/") + "/",
        first_thumb=first_thumb,
        schema_json=json.dumps(schema, indent=2),
        site_root=config.site_root,
        total=len(pages_js),
        pdf_url=pdf_url,
        pages_json=json.dumps(pages_js, ensure_ascii=False),
        pdf_url_json=json.dumps(pdf_url),
        keywords_json=json.dumps(keywords_list),
    )


# ═══════════════════════════════════════════════════════════════
# Entry point — wires everything, then walks
# ═══════════════════════════════════════════════════════════════


@dataclass
class CLIConfig:
    dsn: str
    tenant_slug: str
    root: Path
    media_root: Path
    base_url: str
    site_root: str
    template_dir: Path
    no_recurse: bool
    dry_run: bool


def run(cfg: CLIConfig) -> None:
    conn = psycopg.connect(cfg.dsn, autocommit=True)
    try:
        repo = Repository(conn)
        seo_repo = SeoRepository(conn)

        tenant = repo.get_tenant_by_slug(cfg.tenant_slug)
        if tenant is None:
            print(f"ERROR: no tenant with slug={cfg.tenant_slug!r}", file=sys.stderr)
            sys.exit(1)

        templates = TemplateRegistry(cfg.template_dir)
        resolver = PathResolver(media_root=cfg.media_root, site_root=cfg.site_root)
        doc_index = DocumentIndex(repo, seo_repo, tenant.id)

        walk_config = WalkConfig(
            site_root=cfg.site_root,
            dry_run=cfg.dry_run,
            templates=templates,
            resolver=resolver,
            doc_index=doc_index,
        )

        generate_for_directory(
            directory=cfg.root,
            base_url=cfg.base_url,
            config=walk_config,
            recurse=not cfg.no_recurse,
        )
    finally:
        conn.close()


def main() -> None:
    p = argparse.ArgumentParser(
        description="Generate static HTML for every directory from the document registry DB.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    p.add_argument(
        "--dsn",
        default=os.environ.get("DATABASE_URL", ""),
        help="PostgreSQL DSN (or set DATABASE_URL env var)",
    )
    p.add_argument("--tenant-slug", required=True, dest="tenant_slug")
    p.add_argument(
        "--root", required=True,
        help="Filesystem root to walk (e.g. /srv/media/thedailydialectics/pdfs)",
    )
    p.add_argument(
        "--media-root", required=True, dest="media_root",
        help="Top-level media root for path→URL conversion",
    )
    p.add_argument(
        "--base-url", required=True, dest="base_url",
        help="Public URL for --root (e.g. https://thedailydialectics.com/pdfs)",
    )
    p.add_argument(
        "--site-root",
        default=SITE_ROOT_DEFAULT, dest="site_root",
    )
    p.add_argument(
        "--template-dir",
        default=None, dest="template_dir",
        help="Path to Jinja2 templates (default: ./templates/ next to this script)",
    )
    p.add_argument("--no-recurse", action="store_true", dest="no_recurse")
    p.add_argument("--dry-run", action="store_true", dest="dry_run")
    args = p.parse_args()

    if not args.dsn:
        print("ERROR: --dsn or DATABASE_URL required", file=sys.stderr)
        sys.exit(1)

    template_dir = Path(args.template_dir) if args.template_dir else (
        Path(__file__).resolve().parent / "templates"
    )

    cfg = CLIConfig(
        dsn=args.dsn,
        tenant_slug=args.tenant_slug,
        root=Path(args.root).resolve(),
        media_root=Path(args.media_root).resolve(),
        base_url=args.base_url.rstrip("/"),
        site_root=args.site_root.rstrip("/"),
        template_dir=template_dir,
        no_recurse=args.no_recurse,
        dry_run=args.dry_run,
    )

    if not cfg.root.is_dir():
        print(f"ERROR: {cfg.root} is not a directory", file=sys.stderr)
        sys.exit(1)

    run(cfg)


if __name__ == "__main__":
    main()



from .imports import *
from .constants import *
def get_meta_info(analyzed_data,page_dir,metadata_path):
    image_path = analyzed_data.get("image_path")
    summary = analyzed_data.get("summary", "")
    keywords = analyzed_data.get("keywords", {})
    dirname = os.path.dirname(page_dir)
    dirbase = os.path.basename(dirname)
    parent_dirname = os.path.dirname(dirname)
    parent_dirbase = os.path.basename(parent_dirname)
    pdf_title = parent_dirbase.replace('_','-').replace('_','-').replace('--','-')
    page_title = f"{pdf_title}-{dirbase}"
    alt = f"{pdf_title}-{dirbase} pdf Image"
    capt = summary[:67] if len(summary) >67 else summary
    caption = f"{alt} {capt}..." 
    
    href = path_to_url(image_path)
    meta_info = get_page_data(page_title,
                href,
                summary,
                keywords,
                href,
                alt=alt,
                caption=caption)
    safe_dump_to_json(file_path=str(metadata_path), data=meta_info)
    return meta_info



from .imports import URL_ROOT,generate_meta_tags
from .utils import create_bread_crumbs
import json
def build_image_html(meta: dict, href: str,thumbnail_url=None,text=None,  path=None) -> str:
    
    meta['variants'] = ['https://thedailydialectics.com']
    og    = meta.get("og", {}) or {}
    tw    = meta.get("twitter", {}) or {}
    other = meta.get("other", {}) or {}
    art   = og.get("article", {}) or {}
    description   = meta.get("description", {}) or {}
    domain = tw.get("domain", "")

    domain = tw.get("domain", "")
    if isinstance(domain, list):
        domain = domain[0] if domain else ""

    tag_metas = "".join(
        f'  <meta property="article:tag" content="{t}">\n'
        for t in (art.get("tag") or [])
    )
    article_metas = "".join(
        f'  <meta property="article:{k}" content="{v}">\n'
        for k, v in art.items()
        if k != "tag" and v
    )
    other_metas = "".join(
        f'  <meta name="{k}" content="{v}">\n'
        for k, v in other.items()
        if v and k not in ("alternate", "geo", "hreflang") and not isinstance(v, dict)
    )
    title = meta.get("title", "")
    breadcrumbs = create_bread_crumbs(path) if path else ""
    description = meta.get("description", "")
    keywords    = meta.get("keywords", "")
    canonical   = meta.get("canonical", href)
    schema = {
        "@context":    "https://schema.org",
        "@type":       "CreativeWork",
        "name":        title,
        "description": description,
        "keywords":    keywords,
        "url":         href,
        "thumbnailUrl": thumbnail_url,
        "fileFormat":  "application/pdf",
        "license":     "CC BY-SA 4.0",
        "creditText":  "Created by thedailydialectics for educational purposes",
    }
    first_thumb = thumbnail_url

 
    tags_html   = "".join(
        f'<span class="tag">{kw.strip()}</span>'
        for kw in (meta.get("keywords") or [])
    )
    meta_tags = generate_meta_tags(meta, base_url='https://thedailydialectics.com', json_path=None)
    return f"""<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="{other.get('charset', 'UTF-8')}">
  <meta name="viewport" content="{other.get('viewport', 'width=device-width,initial-scale=1')}">
  <meta name="theme-color" content="{other.get('theme_color', '#FFFFFF')}">
  <meta name="color-scheme" content="{other.get('color_scheme', 'light')}">
  {meta_tags}
  <script type="application/ld+json">{json.dumps(schema, indent=2)}</script>

  <style>
    *, *::before, *::after {{ box-sizing: border-box; margin: 0; padding: 0; }}
    :root {{
      --bg: #0d0d0d; --surface: #1a1a1a; --surface2: #222;
      --text: #e0e0e0; --muted: #888; --accent: #7aaeff;
    }}
    body {{ font-family: system-ui, sans-serif; background: var(--bg); color: var(--text);
            display: flex; flex-direction: column; min-height: 100vh; }}

    .topbar {{ display: flex; align-items: center; gap: .6rem; padding: .5rem .8rem;
               background: var(--surface); border-bottom: 1px solid #333; flex-wrap: wrap; }}
    .topbar a.home {{ color: var(--muted); font-size: .8rem; text-decoration: none; }}
    .topbar a.home:hover {{ color: var(--text); }}
    .topbar h1 {{ font-size: .95rem; font-weight: 600; flex: 1;
                  overflow: hidden; text-overflow: ellipsis; white-space: nowrap; }}

    .abstract-bar {{ background: #111; border-bottom: 1px solid #2a2a2a;
                     padding: .5rem 1rem; font-size: .8rem; color: #999;
                     display: flex; align-items: flex-start; gap: .6rem; }}
    .abstract-bar p {{ flex: 1; line-height: 1.5; }}
    .tags {{ display: flex; flex-wrap: wrap; gap: .25rem; margin-top: .3rem; }}
    .card-tag {{ background: #1e1e1e; border: 1px solid #2e2e2e; border-radius: 3px;
                  padding: .1rem .4rem; font-size: .68rem; color: #666; }}
    .abstract-toggle {{ cursor: pointer; color: var(--muted); font-size: .75rem;
                         white-space: nowrap; user-select: none; }}
    .abstract-toggle:hover {{ color: var(--text); }}

    nav.breadcrumb {{ font-size: .8rem; color: #888; padding: .5rem 1rem;
                      background: var(--surface2); border-bottom: 1px solid #2a2a2a; }}
    nav.breadcrumb a {{ color: #aaa; text-decoration: none; }}
    nav.breadcrumb a:hover {{ color: #fff; }}

    .content {{ flex: 1; display: flex; justify-content: center;
                align-items: flex-start; padding: 1.5rem 1rem; }}
    .image-wrap {{ max-width: 860px; width: 100%; }}
    .image-wrap img {{ width: 100%; height: auto; display: block;
                       border-radius: 4px; border: 1px solid #2a2a2a; }}
    .image-wrap .caption {{ font-size: .8rem; color: var(--muted);
                             margin-top: .5rem; line-height: 1.5; }}

    .page-text {{ max-width: 860px; width: 100%; margin: 1.5rem auto;
                  padding: 0 1rem; }}
    .page-text pre {{ white-space: pre-wrap; font-size: .85rem; line-height: 1.6;
                      color: #ccc; font-family: 'Courier New', monospace;
                      background: var(--surface); padding: 1rem;
                      border-radius: 4px; border: 1px solid #2a2a2a; }}
  </style>
</head>
<body>

<div class="topbar">
  <a class="home" href="https://thedailydialectics.com">← Home</a>
  <span style="color:#444">/</span>
  <h1>{title}</h1>
</div>

<div class="abstract-bar" id="abstract-bar">
  <div style="flex:1">
    <p>{description[:300]}</p>
    <div class="tags">{tags_html}</div>
  </div>
  <span class="abstract-toggle" onclick="
    const c=this.previousElementSibling;
    const hidden=c.style.display==='none';
    c.style.display=hidden?'':'none';
    this.textContent=hidden?'▲ hide':'▼ show info';
  ">▲ hide</span>
</div>

{breadcrumbs}

<div class="content">
  <div class="image-wrap">
    <a href="{href}">
      <img src="{thumbnail_url}"
           alt="{og.get('image_alt', title)}"
           loading="lazy">
    </a>
    <p class="caption">{description[:200]}</p>
  </div>
</div>

{'<div class="page-text"><pre>' + text + '</pre></div>' if text else ''}

</body>
</html>"""




from .imports import *
from .template_registry import *
from .helpers import *
# ═══════════════════════════════════════════════════════════════
# Rendering — templates + context
# ═══════════════════════════════════════════════════════════════
def get_keyword_list(kw):
    if kw and isinstance(kw,list):
        return kw
    if kw and isinstance(kw,str):
        return [eatAll(key,' ') for key in kw.split(',')]
    if kw:
        return list(kw)
    return []

def _find_pdf_in_directory(directory: Path) -> str | None:
    """
    Find the only .pdf file in a directory.
    Returns the relative path from directory root or None if not found.
    """
    pdfs = list(directory.glob("*.pdf"))
    if pdfs:
        return pdfs[0].name  # Just the filename, not full path
    return None


def render_viewer(
    templates: TemplateRegistry,
    view: DocumentView,
    site_root: str,
    document_dir: Path,
) -> str:
    """
    Render the PDF document viewer (index.html for the document).
    
    This is a single-page viewer that displays all pages.
    Automatically finds the PDF in the document directory.
    """
    # Find PDF in the document directory
    pdf_filename = _find_pdf_in_directory(document_dir)
    if not pdf_filename:
        # If no PDF found, this is a problem—but continue with empty
        pdf_filename = ""
    
    pages_js = [
        {
            "n": pv.page_number,
            "thumb": pv.image_url or "",
            "text": pv.text,
            "alt": pv.alt.split(" | ")[0] if pv.alt else f"Page {pv.page_number}",
        }
        for pv in view.pages
    ]
    keywords_list = [
        kw.strip()
        for kw in get_keyword_list(view.keywords)
        if kw.strip() and len(kw.strip()) > 2
    ][:12]
    schema = {
        "@context": "https://schema.org",
        "@type": "CreativeWork",
        "name": view.title,
        "description": clean_text(view.description, 300),
        "keywords": view.keywords,
        "url": "",  # Will be relative/internal
        "thumbnailUrl": view.thumbnail_url or "",
        "fileFormat": "application/pdf",
    }

    return templates.render(
        "viewer",
        title=view.title,
        description=clean_text(view.description, 300),
        keywords=view.keywords,
        canonical_url=view.canonical_url,
        first_thumb=pages_js[0]["thumb"] if pages_js else "",
        schema_json=json.dumps(schema, indent=2),
        site_root=site_root,
        total=len(view.pages),
        pages_json=json.dumps(pages_js, ensure_ascii=False),
        pdf_filename_json=json.dumps(pdf_filename),
        keywords_json=json.dumps(keywords_list),
    )


def render_page_seo(
    templates: TemplateRegistry,
    page_title: str,
    description: str,
    keywords: str,
    alt_text: str,
    image_url: str | None,
    ocr_text: str,
    page_num: int,
    document_title: str,
    document_keywords: str,
    canonical_url: str,
    og_image: str | None,
    page_url: str,
    document_url: str,
    site_root: str,
    schema_data: dict | None = None,
) -> str:
    """
    Render an individual page SEO page.
    
    This page:
    - Has full SEO metadata for this specific page
    - Displays the page image and OCR text
    - Links back to the document viewer
    - Can be indexed and linked directly
    """
    
    # Build schema for this page
    schema = {
        "@context": "https://schema.org",
        "@type": "ImageObject",
        "name": page_title,
        "description": clean_text(description, 300),
        "url": page_url,
        "image": image_url or "",
        "isPartOf": {
            "@type": "CreativeWork",
            "name": document_title,
            "url": document_url,
        },
    }
    
    # Merge with any provided schema data
    if schema_data:
        schema.update(schema_data)
    
    # Extract keywords for JSON
    keywords_list = [
        kw.strip()
        for kw in get_keyword_list(keywords)
        if kw.strip() and len(kw.strip()) > 2
    ][:12]
    
    return templates.render(
        "page_seo",
        title=page_title,
        description=clean_text(description, 300),
        keywords=keywords,
        keywords_list=keywords_list,
        alt_text=alt_text,
        image_url=image_url or "",
        og_image=og_image or image_url or "",
        ocr_text=ocr_text,
        page_num=page_num,
        document_title=document_title,
        document_keywords=document_keywords,
        canonical_url=canonical_url,
        document_url=document_url,
        site_root=site_root,
        schema_json=json.dumps(schema, indent=2),
    )


def render_gallery(
    templates: TemplateRegistry,
    cards: list[GalleryCard],
    title: str,
    base_url: str,
    site_root: str,
) -> str:
    canonical = base_url.rstrip("/") + "/"
    card_dicts = [
        {
            "title": c.title,
            "description": c.description,
            "image_url": c.image_url,
            "href": c.href,
            "tags": c.tags,
            "page_count": c.page_count,
        }
        for c in cards
    ]
    return templates.render(
        "gallery",
        page_title=title,
        canonical_url=canonical,
        breadcrumbs=breadcrumbs_html(base_url, site_root),
        heading=title,
        card_count=len(cards),
        cards=card_dicts,
    )


def render_image_page(
    templates: TemplateRegistry,
    meta: dict[str, Any],
    base_url: str,
    site_root: str,
    directory: Path,
    resolver: PathResolver,
) -> str:
    title = meta.get("title") or humanize(directory.name)
    alt = meta.get("alt") or title
    description = clean_text(
        meta.get("longdesc") or meta.get("caption") or "", 300
    )
    keywords = meta.get("keywords_str", "")
    img_url = (
        meta.get("schema", {}).get("url")
        or meta.get("social_meta", {}).get("og:image")
    )
    # Filesystem fallback for image
    if not img_url:
        img_url = first_image_url(directory, resolver)

    keyword_tags = [kw.strip() for kw in get_keyword_list(keywords) if kw.strip()]

    return templates.render(
        "image_page",
        title=title,
        alt=alt,
        description=description,
        keywords=keywords,
        keyword_tags=keyword_tags,
        img_url=img_url,
        og_image=img_url or "",
        schema_json=json.dumps(meta.get("schema", {}), indent=2),
        license=meta.get("license", ""),
        attribution=meta.get("attribution", ""),
        canonical_url=base_url.rstrip("/") + "/",
        breadcrumbs=breadcrumbs_html(base_url, site_root),
    )



import json


def esc(val):
    if val is None:
        return ""
    return str(val).replace('"', "&quot;")


def tag(name, content):
    if not content:
        return ""
    return f'<meta name="{name}" content="{esc(content)}">'


def prop(name, content):
    if not content:
        return ""
    return f'<meta property="{name}" content="{esc(content)}">'


def build_meta(meta: dict) -> str:
    lines = []

    title = meta.get("title")
    description = meta.get("description")
    keywords = meta.get("keywords")
    canonical = meta.get("canonical")
    image = meta.get("thumbnail_link") or meta.get("og", {}).get("image")

    # BASIC
    if title:
        lines.append(f"<title>{esc(title)}</title>")

    lines.append(tag("description", description))
    lines.append(tag("keywords", keywords))

    # canonical + favicon
    if canonical:
        lines.append(f'<link rel="canonical" href="{esc(canonical)}">')

    if image:
        lines.append(f'<link rel="icon" href="{esc(image)}">')

    # OG
    og = meta.get("og", {})
    lines.append(prop("og:title", og.get("title") or title))
    lines.append(prop("og:description", og.get("description") or description))
    lines.append(prop("og:url", og.get("url") or canonical))
    lines.append(prop("og:image", og.get("image") or image))
    lines.append(prop("og:type", og.get("type", "article")))

    # ARTICLE
    article = og.get("article", {})
    if article.get("published_time"):
        lines.append(prop("article:published_time", article["published_time"]))

    for t in article.get("tag", []):
        lines.append(prop("article:tag", t))

    # TWITTER
    twitter = meta.get("twitter", {})
    lines.append(tag("twitter:card", "summary_large_image"))
    lines.append(tag("twitter:title", twitter.get("title") or title))
    lines.append(tag("twitter:description", twitter.get("description") or description))
    lines.append(tag("twitter:image", twitter.get("image") or image))

    # STRUCTURED DATA
    structured = {
        "@context": "https://schema.org",
        "@type": "Article",
        "headline": title,
        "description": description,
        "image": image,
    }

    lines.append(
        '<script type="application/ld+json">' +
        json.dumps(structured) +
        '</script>'
    )

    return "\n".join([l for l in lines if l])



#!/usr/bin/env python3
from src import *
"""
generate_index_db.py — Static HTML generation from the document registry DB.

Walks the filesystem tree for directory structure, queries the DB for content,
and renders through Jinja2 templates. Every directory gets an index.html:
  - Leaf dirs matching a DB document → viewer page
  - Leaf dirs with info.json (no DB match) → image page
  - Branch dirs with children → gallery index

Design:
  - Filesystem provides structure (which directories exist).
  - DB provides content (titles, descriptions, SEO, page data).
  - Jinja2 templates own all HTML. No f-string markup in Python.
  - Templates are loaded from a registry, not discovered at runtime.
  - All config is explicit. No smart defaults.

Usage:
    python generate_index_db.py \\
        --dsn           "postgresql://user:pass@localhost:5432/tdd_docs" \\
        --tenant-slug   default \\
        --root          /srv/media/thedailydialectics/pdfs \\
        --media-root    /srv/media/thedailydialectics \\
        --base-url      https://thedailydialectics.com/pdfs \\
        --site-root     https://thedailydialectics.com

    # Single directory:
    python generate_index_db.py \\
        --dsn           "postgresql://user:pass@localhost:5432/tdd_docs" \\
        --tenant-slug   default \\
        --root          /srv/media/thedailydialectics/pdfs/wipow/a197278 \\
        --media-root    /srv/media/thedailydialectics \\
        --base-url      https://thedailydialectics.com/pdfs/wipow/a197278 \\
        --site-root     https://thedailydialectics.com \\
        --no-recurse --dry-run
"""














from abstract_webtools import *
from .imports import *
from abstract_react import get_meta_info
def get_parsed_url(domain, **kwargs):
    parsed_url = dict(kwargs)
    post_variants = []
    # http / www
    http_www = get_http_www(domain)
    parsed_url.update(http_www)
    http = http_www.get('http')
    # basic domain pieces
    domain_paths = get_domain_paths(domain, http=http)
    if 'path' not in parsed_url:
        parsed_url['path']=[]
    parsed_url['path']+=domain_paths
    domain_name_ext = get_domain_name_ext(domain, http=http)
    parsed_url.update(domain_name_ext)

    domain_name = parsed_url.get('name',"") or ""
    domain = parsed_url.get('domain',"") or ""
        # tokenization
    tokenized_domain = tokenize_domain(domain)
    parsed_url["tokenized_domain"] = tokenized_domain
    app_name = " ".join(tokenized_domain)
    parsed_url["app_name"] = app_name
    # author / "i_url"
    parsed_url["author"] = f"@{domain_name.lower()}"
    parsed_url["i_url"] = f"{domain_name}://"

    # combine with domain
    # compute final title
    title = get_title(parsed_url)
    post_variants=[title,app_name,domain]
    variants = title_variants_from_domain(domain)
    base_variants = list(set([variant for variant in variants if variant not in post_variants]))
    # update the organized variants
    parsed_url["title_variants"] = get_all_title_variants(variants=base_variants,page=title,name=app_name,domain=domain)

    parsed_url["title"] = pad_or_trim(
        "title",
        string=title,
        title_variants=parsed_url["title_variants"],
        page=title,
        domain=domain,
        name = app_name
    )
    # get keywords
    keywords_info = get_keywords(parsed_url,page=title,domain=domain,name = app_name)
    parsed_url.update(keywords_info)
    keywords = parsed_url.get("keywords", [])
    # FINAL: longest→shortest list with TITLE first, DOMAIN second
    domain = parsed_url.get("domain")
    if domain:
        final_variants = [title, ]
        # remove title/domain from pool
        pool = set(keywords + variants)
        pool.discard(title)
        pool.discard(parsed_url.get("domain"))

        # sort longest → shortest
        final_variants += sort_longest_first(pool)

        parsed_url["title_variants"] = final_variants
   
    return parsed_url
def init_page_data():
    page_data = {}
    variants=VARIANTS,
    title_variants=VARIANTS,
    site_name=SITE_NAME,
    domain=DOMAIN,
    root_url=ROOT_URL
    page_data = getInfo(domain=DOMAIN)
    page_data = get_parsed_url(**page_data)
    page_data['domain']=domain
    page_data['variants']=variants
    page_data['site'] = root_url
    page_data['site_name'] = site_name
    page_data['creator'] = site_name
    page_data['author'] = f'@{site_name}'
    
    return page_data

def get_page_data(
    title,
    href,
    summary,
    keywords,
    thumbnail_url,
    alt=None,
    caption=None,
    ):
    page_data = init_page_data()
    page_data["title"] = title
    page_data['href'] = href
    page_data["page_url"] = href
    page_data['share_url'] = href
    page_data['thumbnail_link'] = thumbnail_url
    page_data['description'] = summary
    page_data['keywords'] = keywords
    page_data['keywords_str']= ','.join(keywords)
    page_data["alt"] = alt or title
    page_data["caption"] = caption or summary
    if href == None:
        return {}
    return get_meta_info(info=page_data,base_image_url=ROOT_URL,base_image_dir=TDD_MEDIA_ROOT_DIR)
##    page_data["alt"] = alt or title
##    page_data["caption"] = caption or summary
##    page_data["schema"] = meta.get("og", {})
##    page_data["schema"]['site_name'] = SITE_NAME
##    page_data["social_meta"] = meta.get("twitter", {})
##    page_data["other"] = meta.get("other")
##    page_data['meta']= meta
##    return page_data



from __future__ import annotations

from jinja2 import Environment, FileSystemLoader
import os

SRC_ROOT = "/var/www/ABSTRACT_ENDEAVORS/scripts/python/generat_htmls/src"

def get_jinja_env() -> Environment:
    return Environment(
        loader=FileSystemLoader(SRC_ROOT),
        autoescape=False,
        trim_blocks=True,
        lstrip_blocks=True,
    )



import os
from concurrent.futures import ThreadPoolExecutor, as_completed

from abstract_utilities import *
from pdf_renderer import *
from processing import process_page_info
from pathlib import Path
from datetime import datetime
import platform

PDF_DIRECTORY = "/srv/media/thedailydialectics/pdfs"
MAX_WORKERS = min(8, (os.cpu_count() or 4))


def info_in_with_txt_in_page_dir(txt_path):
    if os.path.isfile(txt_path):
        try:
            content = read_from_file(txt_path)
            file_parts = get_file_parts(txt_path)
            dirname = file_parts.get('dirname')
            info_path = os.path.join(dirname,'info.json')
            if os.path.isfile(info_path):
               return True 
        except Exception as e:
            print(f"{e}")
    return False
def get_txts_needing_info(pdf_directory: str) -> list[str]:
    """
    Find page image paths whose sibling text.txt does not exist.
    """
    dirs, all_txts = get_files_and_dirs(pdf_directory, allowed_exts=".txt")
    needs_infos = [txt_path for txt_path in all_txts if info_in_with_txt_in_page_dir(txt_path)]
    input(f"{len(needs_infos)} page infos needed")
    return list(reversed(needs_infos))


def safe_process_page_info(txt_path: str) -> tuple[str, bool, str | None]:
    """
    Wrap process_image so one failure does not kill the whole batch.
    """
    try:
        process_page_info(txt_path)
        return image_path, True, None
    except Exception as exc:
        return image_path, False, str(exc)


def main() -> None:
    needs_infos = get_txts_needing_info(PDF_DIRECTORY)

    if not needs_infos:
        print("No images require text extraction.")
        return

    print(f"Processing {len(needs_infos)} images with {MAX_WORKERS} threads...")

    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        futures = {
            executor.submit(safe_process_page_info, txt_path): txt_path
            for txt_path in needs_infos
        }

        for future in as_completed(futures):
            info_path, success, error = future.result()
            if success:
                print(f"[OK] {info_path}")
            else:
                print(f"[FAIL] {info_path} :: {error}")


if __name__ == "__main__":
    main()


    
    




SITE_ROOT_DEFAULT = "https://thedailydialectics.com"

SKIP_DIRS = frozenset({
    "text", "pages", "pdf_pages",
    "preprocessed_images", "preprocessed_text",
    "node_modules", ".git", "__pycache__",
})

IMAGE_EXTS = frozenset({".webp", ".jpg", ".jpeg", ".png", ".gif"})




from .init_imports import *
from .constants import TDD_ROOT_URL,TDD_MEDIA_ROOT_DIR,TDD_ROOT_DIR
from .utils import getPath
def url_to_path(
    url: str,
    media_root: Path or str=None,
    site_root: str=None
    ) -> Optional[Path]:
    site_root = site_root or TDD_ROOT_URL
    if not url or site_root not in url:
        return None
    media_root = media_root or TDD_MEDIA_ROOT_DIR
    rel = url.replace(site_root, "").lstrip("/")
    return getPath(media_root) / rel
def get_rel(string,root):
    if root and root in string:
        return string.replace(root, "").lstrip("/")

def get_rel_path_to_media(path):
    input(path)
    return get_rel(path,TDD_MEDIA_ROOT_DIR)
def get_rel_path_to_root(path):
    return get_rel(path,TDD_ROOT_DIR)
def get_rel_tdd_path(path,root = None):
    rel_path = None
    if not root:
        rel_path = get_rel(path,root)
    if not rel_path:
        rel_path = get_rel_path_to_media(path)
    if not rel_path:
        rel_path = get_rel_path_to_root(path)
    return rel_path
def path_to_url(
    path: Path,
    path_root: str or Path = None,
    site_root: str=None
    ) -> str:
    site_root = site_root or TDD_ROOT_URL
    rel_path = get_rel_tdd_path(path,path_root)
    if rel_path:
        return f"{site_root}/{rel_path}"


def find_correct_path(
    broken: Path,
    media_root: Path
    ) -> Optional[Path]:
    target = broken.name
    if not target:
        return None
    matches = list(media_root.rglob(target))
    if len(matches) == 1:
        return matches[0]
    parent_name = broken.parent.name
    for m in matches:
        if m.parent.name == parent_name:
            return m
    return None


def verified_url(
    url: str,
    media_root: Path,
    site_root: str
    ) -> Optional[str]:
    if not url:
        return None
    expected = url_to_path(url, media_root, site_root)
    if expected is None:
        return url
    if expected.exists():
        return url
    correct = find_correct_path(expected, media_root)
    return path_to_url(correct, media_root, site_root) if correct else None




"""
repository.py — Data access layer for the document registry.

Uses psycopg 3 (import psycopg, not psycopg2).

Design decisions:
  - Connection is injected, never acquired internally.
  - Return types are dataclasses, never raw tuples.
  - Every public method is one logical operation.
  - Connection should be created with autocommit=True.
  - No module-level state. Wire a Repository instance at your composition root.
"""




# ── Repository ───────────────────────────────────────────────
from .imports import *
from .classes import *
from .row_mappers import (
    _to_page,
    _to_analysis,
    _to_run,
    _to_tag,
    _to_document,
    _to_tenant
    )
class Repository:
    """
    All database operations for the document registry.

    Usage:
        conn = psycopg.connect(dsn, autocommit=True)
        repo = Repository(conn)
        doc  = repo.upsert_document(tenant_id, base_dir)
    """

    def __init__(self, conn: psycopg.Connection) -> None:
        self._conn = conn

    def _cursor(self):
        return self._conn.cursor(row_factory=dict_row)

    # ── tenants ──

    def insert_tenant(self, name: str, slug: str) -> Tenant:
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO tenants (name, slug)
                VALUES (%s, %s)
                RETURNING *
            """, (name, slug))
            return _to_tenant(cur.fetchone())

    def get_tenant(self, tenant_id: str) -> Tenant | None:
        with self._cursor() as cur:
            cur.execute("SELECT * FROM tenants WHERE id = %s", (tenant_id,))
            row = cur.fetchone()
            return _to_tenant(row) if row else None

    def get_tenant_by_slug(self, slug: str) -> Tenant | None:
        with self._cursor() as cur:
            cur.execute("SELECT * FROM tenants WHERE slug = %s", (slug,))
            row = cur.fetchone()
            return _to_tenant(row) if row else None

    # ── documents ──

    def insert_document(
        self,
        tenant_id: str,
        slug: str,
        base_path: str,
        pdf_path: str,
        discriminator: str = "",
    ) -> Document:
        doc_id = identity_hash(slug, discriminator)
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO documents (tenant_id, doc_id, slug, base_path, pdf_path)
                VALUES (%s, %s, %s, %s, %s)
                RETURNING *
            """, (tenant_id, doc_id, slug, base_path, pdf_path))
            return _to_document(cur.fetchone())

    def upsert_document(
        self,
        tenant_id: str,
        slug: str,
        base_path: str,
        pdf_path: str,
        discriminator: str = "",
    ) -> Document:
        doc_id = identity_hash(slug, discriminator)
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO documents (tenant_id, doc_id, slug, base_path, pdf_path)
                VALUES (%s, %s, %s, %s, %s)
                ON CONFLICT (tenant_id, doc_id) DO UPDATE
                    SET base_path  = EXCLUDED.base_path,
                        pdf_path   = EXCLUDED.pdf_path,
                        updated_at = now()
                RETURNING *
            """, (tenant_id, doc_id, slug, base_path, pdf_path))
            return _to_document(cur.fetchone())

    def get_document(self, document_id: int) -> Document | None:
        with self._cursor() as cur:
            cur.execute("SELECT * FROM documents WHERE id = %s", (document_id,))
            row = cur.fetchone()
            return _to_document(row) if row else None

    def get_document_by_hash(self, tenant_id: str, doc_id: str) -> Document | None:
        with self._cursor() as cur:
            cur.execute(
                "SELECT * FROM documents WHERE tenant_id = %s AND doc_id = %s",
                (tenant_id, doc_id),
            )
            row = cur.fetchone()
            return _to_document(row) if row else None

    def list_documents(
        self,
        tenant_id: str,
        status: Status | None = None,
        limit: int = 100,
        offset: int = 0,
    ) -> list[Document]:
        clauses = ["tenant_id = %s"]
        params: list[Any] = [tenant_id]
        if status is not None:
            clauses.append("status = %s")
            params.append(status.value)
        params.extend([limit, offset])

        with self._cursor() as cur:
            cur.execute(f"""
                SELECT * FROM documents
                WHERE {" AND ".join(clauses)}
                ORDER BY created_at DESC
                LIMIT %s OFFSET %s
            """, params)
            return [_to_document(r) for r in cur.fetchall()]

    def update_document_status(self, document_id: int, status: Status) -> Document:
        with self._cursor() as cur:
            cur.execute("""
                UPDATE documents SET status = %s WHERE id = %s RETURNING *
            """, (status.value, document_id))
            row = cur.fetchone()
            if row is None:
                raise LookupError(f"document {document_id} not found")
            return _to_document(row)

    def update_page_count(self, document_id: int, page_count: int) -> Document:
        with self._cursor() as cur:
            cur.execute("""
                UPDATE documents SET page_count = %s WHERE id = %s RETURNING *
            """, (page_count, document_id))
            row = cur.fetchone()
            if row is None:
                raise LookupError(f"document {document_id} not found")
            return _to_document(row)

    def relocate_document(
        self, tenant_id: str, doc_id: str, new_base_path: str
    ) -> int:
        with self._cursor() as cur:
            cur.execute(
                "SELECT relocate_document(%s, %s, %s)",
                (tenant_id, doc_id, new_base_path),
            )
            row = cur.fetchone()
            return list(row.values())[0]

    def delete_document(self, document_id: int) -> bool:
        with self._cursor() as cur:
            cur.execute(
                "DELETE FROM documents WHERE id = %s RETURNING id", (document_id,)
            )
            return cur.fetchone() is not None

    # ── pages ──

    def insert_page(
        self,
        document_id: int,
        page_number: int,
        image_path: str | None = None,
        text_path: str | None = None,
        info_path: str | None = None,
        metadata_path: str | None = None,
    ) -> Page:
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO pages
                    (document_id, page_number, image_path, text_path, info_path, metadata_path)
                VALUES (%s, %s, %s, %s, %s, %s)
                RETURNING *
            """, (document_id, page_number, image_path, text_path, info_path, metadata_path))
            return _to_page(cur.fetchone())

    def upsert_page(
        self,
        document_id: int,
        page_number: int,
        image_path: str | None = None,
        text_path: str | None = None,
        info_path: str | None = None,
        metadata_path: str | None = None,
    ) -> Page:
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO pages
                    (document_id, page_number, image_path, text_path, info_path, metadata_path)
                VALUES (%s, %s, %s, %s, %s, %s)
                ON CONFLICT (document_id, page_number) DO UPDATE
                    SET image_path    = COALESCE(EXCLUDED.image_path,    pages.image_path),
                        text_path     = COALESCE(EXCLUDED.text_path,     pages.text_path),
                        info_path     = COALESCE(EXCLUDED.info_path,     pages.info_path),
                        metadata_path = COALESCE(EXCLUDED.metadata_path, pages.metadata_path)
                RETURNING *
            """, (document_id, page_number, image_path, text_path, info_path, metadata_path))
            return _to_page(cur.fetchone())

    def get_pages(self, document_id: int) -> list[Page]:
        with self._cursor() as cur:
            cur.execute("""
                SELECT * FROM pages
                WHERE document_id = %s
                ORDER BY page_number
            """, (document_id,))
            return [_to_page(r) for r in cur.fetchall()]

    def bulk_insert_pages(self, document_id: int, pages: list[dict]) -> int:
        if not pages:
            return 0
        with self._cursor() as cur:
            cur.executemany(
                """
                INSERT INTO pages
                    (document_id, page_number, image_path, text_path, info_path, metadata_path)
                VALUES (%s, %s, %s, %s, %s, %s)
                ON CONFLICT (document_id, page_number) DO NOTHING
                """,
                [
                    (
                        document_id,
                        p["page_number"],
                        p.get("image_path"),
                        p.get("text_path"),
                        p.get("info_path"),
                        p.get("metadata_path"),
                    )
                    for p in pages
                ],
            )
            return cur.rowcount

    # ── page analysis ──

    def upsert_analysis(
        self,
        page_id: int,
        analysis_type: str,
        payload: dict[str, Any],
        model_version: str | None = None,
    ) -> PageAnalysis:
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO page_analysis (page_id, analysis_type, payload, model_version)
                VALUES (%s, %s, %s, %s)
                ON CONFLICT (page_id, analysis_type) DO UPDATE
                    SET payload       = EXCLUDED.payload,
                        model_version = EXCLUDED.model_version,
                        created_at    = now()
                RETURNING *
            """, (page_id, analysis_type, Jsonb(payload), model_version))
            return _to_analysis(cur.fetchone())

    def get_analysis(self, page_id: int, analysis_type: str) -> PageAnalysis | None:
        with self._cursor() as cur:
            cur.execute(
                "SELECT * FROM page_analysis WHERE page_id = %s AND analysis_type = %s",
                (page_id, analysis_type),
            )
            row = cur.fetchone()
            return _to_analysis(row) if row else None

    def get_all_analysis(self, page_id: int) -> list[PageAnalysis]:
        with self._cursor() as cur:
            cur.execute(
                "SELECT * FROM page_analysis WHERE page_id = %s ORDER BY analysis_type",
                (page_id,),
            )
            return [_to_analysis(r) for r in cur.fetchall()]

    # ── pipeline runs ──

    def start_pipeline_run(self, document_id: int, stage: str) -> PipelineRun:
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO pipeline_runs (document_id, status, stage)
                VALUES (%s, %s, %s)
                RETURNING *
            """, (document_id, Status.INGESTING.value, stage))
            return _to_run(cur.fetchone())

    def complete_pipeline_run(self, run_id: int) -> PipelineRun:
        with self._cursor() as cur:
            cur.execute("""
                UPDATE pipeline_runs
                SET status = %s, finished_at = now()
                WHERE id = %s
                RETURNING *
            """, (Status.COMPLETE.value, run_id))
            row = cur.fetchone()
            if row is None:
                raise LookupError(f"pipeline run {run_id} not found")
            return _to_run(row)

    def fail_pipeline_run(self, run_id: int, error: str) -> PipelineRun:
        with self._cursor() as cur:
            cur.execute("""
                UPDATE pipeline_runs
                SET status = %s, error_message = %s, finished_at = now()
                WHERE id = %s
                RETURNING *
            """, (Status.FAILED.value, error, run_id))
            row = cur.fetchone()
            if row is None:
                raise LookupError(f"pipeline run {run_id} not found")
            return _to_run(row)

    def get_pipeline_history(self, document_id: int) -> list[PipelineRun]:
        with self._cursor() as cur:
            cur.execute("""
                SELECT * FROM pipeline_runs
                WHERE document_id = %s
                ORDER BY started_at DESC
            """, (document_id,))
            return [_to_run(r) for r in cur.fetchall()]

    # ── tags ──

    def ensure_tag(self, slug: str, label: str) -> Tag:
        with self._cursor() as cur:
            cur.execute("""
                INSERT INTO tags (slug, label)
                VALUES (%s, %s)
                ON CONFLICT (slug) DO UPDATE SET label = EXCLUDED.label
                RETURNING *
            """, (slug, label))
            return _to_tag(cur.fetchone())

    def tag_document(self, document_id: int, tag_slug: str) -> None:
        with self._cursor() as cur:
            cur.execute("SELECT id FROM tags WHERE slug = %s", (tag_slug,))
            row = cur.fetchone()
            if row is None:
                raise LookupError(f"tag '{tag_slug}' not in registry")
            cur.execute("""
                INSERT INTO document_tags (document_id, tag_id)
                VALUES (%s, %s)
                ON CONFLICT DO NOTHING
            """, (document_id, row["id"]))

    def untag_document(self, document_id: int, tag_slug: str) -> None:
        with self._cursor() as cur:
            cur.execute("""
                DELETE FROM document_tags
                WHERE document_id = %s
                  AND tag_id = (SELECT id FROM tags WHERE slug = %s)
            """, (document_id, tag_slug))

    def get_document_tags(self, document_id: int) -> list[Tag]:
        with self._cursor() as cur:
            cur.execute("""
                SELECT t.* FROM tags t
                JOIN document_tags dt ON dt.tag_id = t.id
                WHERE dt.document_id = %s
                ORDER BY t.slug
            """, (document_id,))
            return [_to_tag(r) for r in cur.fetchall()]

    def list_documents_by_tag(
        self, tenant_id: str, tag_slug: str
    ) -> list[Document]:
        with self._cursor() as cur:
            cur.execute("""
                SELECT d.* FROM documents d
                JOIN document_tags dt ON dt.document_id = d.id
                JOIN tags t ON t.id = dt.tag_id
                WHERE d.tenant_id = %s AND t.slug = %s
                ORDER BY d.created_at DESC
            """, (tenant_id, tag_slug))
            return [_to_document(r) for r in cur.fetchall()]

    def update_page_ocr_status(self, page_id: int, ocr_processed: bool) -> Page:
            """Mark a page as OCR processed (or not)."""
            with self._cursor() as cur:
                cur.execute("""
                    UPDATE pages SET ocr_processed = %s WHERE id = %s RETURNING *
                """, (ocr_processed, page_id))
                row = cur.fetchone()
                if row is None:
                    raise LookupError(f"page {page_id} not found")
                return _to_page(row)

    def get_pages_needing_ocr(self, document_id: int) -> list[Page]:
        """Get all pages where ocr_processed = False."""
        with self._cursor() as cur:
            cur.execute("""
                SELECT * FROM pages
                WHERE document_id = %s AND ocr_processed = False
                ORDER BY page_number
            """, (document_id,))
            return [_to_page(r) for r in cur.fetchall()]
    def save_page_content(
        self,
        page_id: int,
        text_content: str,
        analysis: dict,
        metadata: dict,
    ) -> Page:
        """Save page with explicit analysis and metadata schemas."""
        with self._cursor() as cur:
            cur.execute("""
                UPDATE pages SET 
                    text_content = %s,
                    analysis = %s,
                    metadata = %s,
                    ocr_processed = TRUE
                WHERE id = %s
                RETURNING *
            """, (
                text_content,
                Jsonb(analysis),
                Jsonb(metadata),
                page_id
            ))
            return _to_page(cur.fetchone())

    def get_page_content(self, page_id: int) -> dict | None:
        """Get page with typed analysis and metadata."""
        with self._cursor() as cur:
            cur.execute("""
                SELECT id, page_number, text_content, analysis, metadata, ocr_processed
                FROM pages WHERE id = %s
            """, (page_id,))
            return cur.fetchone()
    def ensure_document_pages(self, document_id: int, page_count: int) -> list[Page]:
        """Create page records for all pages in a document if they don't exist."""
        with self._cursor() as cur:
            # Get existing pages
            cur.execute("""
                SELECT page_number FROM pages WHERE document_id = %s
            """, (document_id,))
            existing = {row['page_number'] for row in cur.fetchall()}
            
            # Create missing pages
            pages_to_create = [
                (document_id, i, None, None, None, None)
                for i in range(1, page_count + 1)
                if i not in existing
            ]
            
            if pages_to_create:
                cur.executemany("""
                    INSERT INTO pages
                        (document_id, page_number, image_path, text_path, info_path, metadata_path)
                    VALUES (%s, %s, %s, %s, %s, %s)
                """, pages_to_create)
            
            # Return all pages for this document
            cur.execute("""
                SELECT * FROM pages WHERE document_id = %s ORDER BY page_number
            """, (document_id,))
            return [_to_page(r) for r in cur.fetchall()]
    def mark_page_complete(self, page_id: int) -> Page:
        """Mark a page as fully processed."""
        with self._cursor() as cur:
            cur.execute("""
                UPDATE pages SET page_complete = TRUE WHERE id = %s RETURNING *
            """, (page_id,))
            row = cur.fetchone()
            if row is None:
                raise LookupError(f"page {page_id} not found")
            return _to_page(row)

    def get_incomplete_pages(self, document_id: int) -> list[Page]:
        """Get all pages where page_complete = False."""
        with self._cursor() as cur:
            cur.execute("""
                SELECT * FROM pages
                WHERE document_id = %s AND page_complete = FALSE
                ORDER BY page_number
            """, (document_id,))
            return [_to_page(r) for r in cur.fetchall()]
    def delete_page(self, page_id: int) -> bool:
        """Delete a page record."""
        with self._cursor() as cur:
            cur.execute("DELETE FROM pages WHERE id = %s RETURNING id", (page_id,))
            return cur.fetchone() is not None

    def delete_document(self, document_id: int) -> int:
        """Delete a document and all its pages. Returns count deleted."""
        with self._cursor() as cur:
            cur.execute("DELETE FROM pages WHERE document_id = %s", (document_id,))
            pages_deleted = cur.rowcount
            
            cur.execute("DELETE FROM documents WHERE id = %s", (document_id,))
            
            return pages_deleted
